"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.callIn = callIn;

var _proxyequal = require("proxyequal");

var _cache = require("./cache");

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

var nothing = {
  is: 'PROXY_EQUAL_NOTHING'
};
var emptyArray = [];

function addAffected(callGeneration, affected, object) {
  var key = (0, _proxyequal.getProxyKey)(object);

  if (key.fingerPrint.callGeneration === callGeneration) {
    affected[key.fingerPrint.index].resultAffected.push(key.suffix);
    return true;
  }

  return false;
}

var cycleMap = null;
var userShouldDive = null;
var knownPOD = new WeakSet();

var defaultShouldDiveCheck = function defaultShouldDiveCheck(line, key, object) {
  // default check - is not React owner
  if (key === '_owner' && object.$$typeof) {
    // React-specific: avoid traversing React elements' _owner.
    //  _owner contains circular references
    // and is not needed when comparing the actual elements (and not their owners)
    // .$$typeof and ._store on just reasonable markers of a react element
    return false;
  } // could not ACCESS XMLHttpRequest


  if (global.XMLHttpRequest && line.constructor === global.XMLHttpRequest) {
    return false;
  }

  return true;
};

var shouldDive = function shouldDive(line, key, object) {
  return _typeof(line) === 'object' && // !knownPOD.has(line) &&
  !(0, _proxyequal.isKnownObject)(line) && ((0, _proxyequal.isProxyfied)(line) || userShouldDive(line, key, object, defaultShouldDiveCheck));
};

function forEachIn(obj, iterator) {
  if (Array.isArray(obj)) {
    obj.forEach(function (_, index) {
      return iterator(index);
    });
  }

  Object.keys(obj).forEach(iterator);
}

function deproxifyResult(callGeneration, result, affected, returnPureValue) {
  var deepDive = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : false;
  var unaffectedResult = returnPureValue ? result : nothing;

  if (!result) {
    return result;
  }

  if (_typeof(result) === 'object') {
    if (knownPOD.has(result)) {
      return unaffectedResult;
    }

    var isInProxy = (0, _proxyequal.isProxyfied)(result);

    if (isInProxy) {
      if (addAffected(callGeneration, affected, result)) {
        var preResult = (0, _proxyequal.deproxify)(result);
        return deepDive ? preResult : deproxifyResult(callGeneration, preResult, affected, true, true);
      }
    }

    var sub = Array.isArray(result) ? [] : {};
    var altered = false;

    if (result && Object.getOwnPropertyDescriptor(result, '__proxyequal_scanEnd')) {
      Object.defineProperty(result, '__proxyequal_scanEnd', {
        value: 'here was spread guard',
        configurable: true,
        enumerable: false
      });
    }

    forEachIn(result, function (i) {
      var data = result[i];
      var newResult = data;

      if (data && shouldDive(data, i, result)) {
        if (!cycleMap.has(data)) {
          cycleMap.set(data, nothing);
          newResult = deproxifyResult(callGeneration, data, affected, false, deepDive);
          cycleMap.set(data, newResult); // knownPOD.add(newResult === nothing ? data : newResult);
        } else {
          newResult = cycleMap.get(data);
        }
      }

      if (data && newResult !== nothing && newResult !== data) {
        altered = true;
        sub[i] = newResult;
      } else {
        sub[i] = data;
      }
    });

    if (altered) {
      return sub;
    }

    knownPOD.add(result);
    return unaffectedResult;
  }

  return unaffectedResult;
}

function callIn(that, cache, args, func, memoizationDepth) {
  var proxyMap = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : [];
  var options = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : {};
  var proxies = args.map(function (arg, index) {
    if (arg && _typeof(arg) === 'object') {
      var map = proxyMap[index];

      if (!map) {
        return proxyMap[index] = (0, _proxyequal.proxyState)(arg, {
          callGeneration: proxyMap,
          index: index
        });
      }

      map.reset();
      return map.replaceState(arg);
    }

    return undefined;
  });
  var callArgs = args.map(function (arg, index) {
    return proxies[index] ? proxies[index].state : arg;
  }); //call the real function

  var preResult = func.call.apply(func, [that].concat(_toConsumableArray(callArgs)));
  proxies.forEach(function (proxy) {
    return proxy && proxy.seal();
  });
  var spreadDetected = [];
  var affected = proxies.map(function (proxy) {
    if (proxy) {
      if (proxy.spreadDetected !== false) {
        spreadDetected.push(proxy.spreadDetected);
      }

      var _affected = proxy.affected || emptyArray;

      return {
        useAffected: _toConsumableArray(_affected),
        resultAffected: []
      };
    }

    return undefined;
  });
  cycleMap = new WeakMap();
  userShouldDive = options.deproxifyShouldDive || defaultShouldDiveCheck;
  var result = deproxifyResult(proxyMap, preResult, affected, true);
  var cacheLine = {
    args: args,
    affected: affected,
    result: result,
    callArgs: callArgs
  };

  if (cache.length < memoizationDepth) {
    cache.push(cacheLine);
  } else {
    (0, _cache.updateCacheLine)(cache, 0, cacheLine);
  }

  if (spreadDetected.length > 0) {
    if (process.env.NODE_ENV !== 'production') {
      console.warn('memoize-state: object spread detected in ', func, '. Keys affected: ', spreadDetected.map(function (key) {
        return key ? key : 'root';
      }), '. This is no-op.');
    }
  }

  return result;
}